(* O(i) *)
let log_choose n i = 
  let fn = float n
  and fi = float i in
  let ret = ref 0. in
  for j = 1 to i do
    ret := !ret +. log ((float (n - j +1)) /. (float j))
  done;!ret

(* O(n) *)
let rec log_factorial n = log (float n) +. log_factorial (n-1)

let bernoulli_number = [|1.; ~-. 1. /. 2.; 
	    1. /. 6.; 0.; 
	    ~-. 1. /. 30.; 0.;
	    1. /. 42.; 0.; 
	    ~-. 1. /. 30.; 0.; 
	    5. /. 66.; 0.; 
	    ~-.691. /. 2730.; 0.; 
	    7. /. 6.; 0.; 
	    ~-.3617. /. 510.; 0.; 
	    43867. /. 798.; 0.; 
	    ~-. 174611. /. 330.; 0.; 
	    854513. /. 138. |]
let bernoulli_length = 22
    
let half_log_two_pi = 0.5 *. log (2. *. 3.14159)
    
(*Calculate n!  
   Approximation = Stirling's series*)
let approx_log_factorial n =
  if n = 0 
  then 0.
  else
    let z = float (n+1) in
    let ret = ref (half_log_two_pi +. (z -. 0.5) *. log(z) -. z) in
    for i = 1 to bernoulli_length / 2 do
      ret := !ret +. bernoulli_number.(i * 2) /. float (2 * i * (2 * i - 1)) /. z ** float (2*i - 1);
    done;
    !ret
      
let approx_log_choose n i =
  approx_log_factorial n -. approx_log_factorial i -. approx_log_factorial (n-i)

let log_catalan n = approx_log_choose (2*n) n -. log (float (n+1))

