let rec monotone_increasing f too_tiny (target : float) min max =
  if max -. min <= too_tiny then max
  else 
    let mid = (max +. min) /. 2. in
    let v = f mid in
    if Mfloat.near too_tiny v target 
    then mid
    else
      if v < target then monotone_increasing f too_tiny target mid max
      else monotone_increasing f too_tiny target min mid  

let rec monotone_decreasing f too_tiny (target : float) min max =
  if max -. min <= too_tiny then min
  else 
    let mid = (max +. min) /. 2. in
    let v = f mid in
    if Mfloat.near too_tiny v target 
    then mid
    else
      if v < target then monotone_decreasing f too_tiny target min mid
      else monotone_decreasing f too_tiny target mid max

let rec monotone_decreasing_int f (target : float) min max =
  if max - min <= 1 then max
  else 
    let mid = (max + min) / 2 in
    let v = f mid in
    if v < target then monotone_decreasing_int f target min mid
    else monotone_decreasing_int f target mid max

(* We are guaranteed: min_val > mid_val < max_val 
Is there a lower point to the left or to the right?

Invariant of the search algorithm: any three points with the middle point having a lower 
value contains the interval.

*)
let rec find_min f too_tiny min mid max = 
  let mid_val = f mid in
  let mid_high = (mid +. max) /. 2. in
  
  if f mid_high < mid_val -. too_tiny
  then find_min f too_tiny mid mid_high max 
  else
    let mid_low = (mid +. min) /. 2. in
    let mid_low_val = f mid_low in
    if mid_low_val < mid_val -. too_tiny
    then find_min f too_tiny min mid_low mid 
    else
      if mid_val > mid_low_val -. too_tiny 
      then mid
      else find_min f too_tiny mid_low mid mid_high 

let rec minimize_convex f too_tiny min max = 
  let mid = (min +. max) /. 2. in
  let min_val = f min
  and mid_val = f mid in
  if min_val < mid_val -. too_tiny 
  then minimize_convex f too_tiny min mid
  else 
    let max_val = f max in
    if max_val < mid_val -. too_tiny
    then minimize_convex f too_tiny mid max
    else 
      find_min f too_tiny min mid max

let rec linesearch_to_min f last_val min max =
  let next = min *. 2. in
  if next < max 
  then
    let next_val = f next in
    if next_val <= fst last_val
    then linesearch_to_min f (next_val,next) next max
    else snd last_val
  else
    snd last_val

let linesearch_to_min f min max =
  linesearch_to_min f ((f 0.),0.) min max

